﻿using Microsoft.ProjectOxford.Face;
using Microsoft.ProjectOxford.Face.Contract;
using Microsoft.VisualBasic;
using MicrosoftProjectOxford;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net.Mail;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace FaceAPIApp
{

    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window {
        string speechAPIKey = ConfigurationManager.AppSettings["SpeechAPIKey"];
        static string faceAPIKey = ConfigurationManager.AppSettings["FaceAPIKey"];

        private string personGroupId = Guid.NewGuid().ToString();

        private MicrophoneRecognitionClient m_micClient;
        private readonly IFaceServiceClient faceServiceClient = new FaceServiceClient(faceAPIKey);

        private List<string> mailList = new List<string>();

        public MainWindow() {
            InitializeComponent();
            Intialize();
        }

        /// <summary>
        /// Init speech api
        /// </summary>
        private async void Intialize()
        {
            m_micClient = SpeechRecognitionServiceFactory.CreateMicrophoneClient(SpeechRecognitionMode.ShortPhrase, "de-de", speechAPIKey);
            m_micClient.OnResponseReceived += this.OnResponseReceivedHandler;
            m_micClient.OnPartialResponseReceived += this.OnPartialResponseReceivedHandler;
            m_micClient.OnConversationError += this.OnConversationErrorHandler;
            m_micClient.OnMicrophoneStatus += this.OnMicrophoneStatus;

            await faceServiceClient.CreatePersonGroupAsync(personGroupId, personGroupId);
            await AutoDetectFaces(@"..\..\..\..\example_pictures\Anna", "s_d@uni.de");
            await AutoDetectFaces(@"..\..\..\..\example_pictures\Bill", "natalie-kurz1@gmx.de");
            await StartTraining();
            BrowseButton.IsEnabled = true;
        }

        /// <summary>
        /// Detect faces from by uploading an image asynchronously
        /// </summary>
        /// <param name="imageFilePath">The file path of the image file</param>
        /// <returns>
        /// A Task object represents the future result of detection.
        /// The result will be an array of Face objects 
        /// </returns>
        private async Task<Face[]> UploadAndDetectFaces(string imageFilePath) {
            try {
                using (Stream imageFileStream = File.OpenRead(imageFilePath)) {
                    var faces = await faceServiceClient.DetectAsync(imageFileStream);
                    return faces;
                }
            }
            catch (Exception) {
                Console.WriteLine("Error: Could not detect faces.");
                return new Face[0];
            }
        }

        /// <summary>
        /// Checks if a person is recognized in image
        /// </summary>
        /// <param name="person"></param>
        /// <param name="faces"></param>
        /// <returns></returns>
        private async void SearchSimilarFaces(Face[] faces) {
            var faceIds = faces.Select(face => face.FaceId).ToArray();
            var results = await faceServiceClient.IdentifyAsync(personGroupId, faceIds, 4);

            foreach (var identifyResult in results)
            {
                if (identifyResult.Candidates.Length != 0)
                {
                    var candidateId = identifyResult.Candidates[0].PersonId;
                    var person = await faceServiceClient.GetPersonAsync(personGroupId, candidateId);
                    mailList.Add(person.Name);
                    this.speechButton.IsEnabled = true;
                }
            }
            this.WriteLine("Found {0} similar faces. Enable Speech API to send Mail!", mailList.Count);
        }

        private async void BrowseButton_Click(object sender, RoutedEventArgs e) {
            var openDlg = new Microsoft.Win32.OpenFileDialog();

            openDlg.Filter = "JPEG Image(*.jpg)|*.jpg";
            bool? result = openDlg.ShowDialog(this);

            if ((bool)result) {
                BitmapImage bitmapSource = DrawImage(openDlg.FileName);
                
                this.WriteLine("Detecting...");
                Face[] faces = await UploadAndDetectFaces(openDlg.FileName);
                this.WriteLine("Detection Finished. {0} face(s) detected", faces.Length);

                if (faces.Length > 0) {
                    DrawDetectedFaceBorders(bitmapSource, faces, Brushes.Red);
                    /*
                    if (faces.Length == 1) {
                        var email = Interaction.InputBox("Please, insert the email address of the shown person", "Email address", "");
                        if (email != "") { //email pattern match ?!
                            var friendFaceIds = faces.Select(face => face.FaceId).ToArray();
                            CreatePersonResult friend = await faceServiceClient.CreatePersonAsync(personGroupId, friendFaceIds, email);
                            WriteLine("await-1");
                            await startTraining();
                            WriteLine("await-12");
                            this.WriteLine("{0} added to known persons ", email);
                        }
                    }
                    else {
                    }
                     */
                    SearchSimilarFaces(faces);
                }
            }
        }

        /// <summary>
        /// Draw face rectangles on image
        /// </summary>
        /// <param name="bitmapSource"></param>
        /// <param name="faces"></param>
        /// <param name="color"></param>
        private void DrawDetectedFaceBorders(BitmapImage bitmapSource, Face[] faces, Brush color) {
            var faceRects = faces.Select(face => face.FaceRectangle).ToArray();
            DrawingVisual visual = new DrawingVisual();
            DrawingContext drawingContext = visual.RenderOpen();

            drawingContext.DrawImage(bitmapSource, new Rect(0, 0, bitmapSource.Width, bitmapSource.Height));

            double dpi = bitmapSource.DpiX;
            double resizeFactor = 96 / dpi;

            foreach (var faceRect in faceRects) {
                drawingContext.DrawRectangle(
                    Brushes.Transparent,
                    new Pen(color, 2),
                    new Rect(faceRect.Left * resizeFactor, faceRect.Top * resizeFactor, faceRect.Width * resizeFactor, faceRect.Height * resizeFactor)
                );
            }

            drawingContext.Close();

            RenderTargetBitmap faceWithRectBitmap = new RenderTargetBitmap(
                (int)(bitmapSource.PixelWidth * resizeFactor), (int)(bitmapSource.PixelHeight * resizeFactor), 96, 96, PixelFormats.Pbgra32);

            faceWithRectBitmap.Render(visual);
            FacePhoto.Source = faceWithRectBitmap;
        }


        private async Task<bool> AutoDetectFaces(string imageDir, string email){
            List<Face> friendFaces = new List<Face>();

            foreach (string imagePath in Directory.GetFiles(imageDir, "*.jpg"))
            {
                Console.WriteLine(imagePath);
                using (Stream s = File.OpenRead(imagePath))
                {
                    var faces = await faceServiceClient.DetectAsync(s);

                    if (faces.Length != 0) {
                        friendFaces.Add(faces[0]);
                    }
                }
            }

            var friendFaceIds = friendFaces.Select(face => face.FaceId).ToArray();
            var friend = await faceServiceClient.CreatePersonAsync(personGroupId, friendFaceIds, email);
            return true;
        }

        private async Task StartTraining()
        {
            // Train the person group
            await faceServiceClient.TrainPersonGroupAsync(personGroupId);

            TrainingStatus trainingStatus = null;
            while (true)
            {
                trainingStatus = await faceServiceClient.GetPersonGroupTrainingStatusAsync(personGroupId);

                if (trainingStatus.Status != "running")
                {
                    break;
                }
                await Task.Delay(1000);
            }
            WriteLine("Training succesful.");
        }

        /// <summary>
        /// Draw an image on the image-frame
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        private BitmapImage DrawImage(string filePath) {

            Uri fileUri = new Uri(filePath);
            BitmapImage bitmapSource = new BitmapImage();

            bitmapSource.BeginInit();
            bitmapSource.CacheOption = BitmapCacheOption.None;
            bitmapSource.UriSource = fileUri;
            bitmapSource.EndInit();

            FacePhoto.Source = bitmapSource;
            return bitmapSource;
        }

        /// <summary>
        /// Handles the Click event of the speechButton control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void SpeechButton_Click(object sender, RoutedEventArgs e)
        {
            this.speechButton.IsEnabled = false;
            m_micClient.StartMicAndRecognition();
        }

        /// <summary>
        ///     Called when a final response is received; 
        /// </summary>
        void OnResponseReceivedHandler(object sender, SpeechResponseEventArgs e)
        {
            var status = e.PhraseResponse.RecognitionStatus;
            if (status == RecognitionStatus.EndOfDictation || status == RecognitionStatus.DictationEndSilenceTimeout)
            {
                this.Dispatcher.Invoke((Action)(() =>
                {
                    this.speechButton.IsEnabled = true;
                }));
            }
            else
            {
                this.WriteLine("** Final NBEST Results:");
                for (int i = 0; i < e.PhraseResponse.Results.Length; i++)
                {
                    this.WriteLine("[{0}] Confidence={1} Text=\"{2}\"",
                                   i, e.PhraseResponse.Results[i].Confidence,
                                   e.PhraseResponse.Results[i].DisplayText);
                }

                RecognizedPhrase[] results = e.PhraseResponse.Results;
                AnalyseSpeechText(results);
            }
        }

        private void AnalyseSpeechText(RecognizedPhrase[] results)
        {
            if (results.Length > 0 && mailList.Count > 0)
            {
                // Get first speech text result
                string speechText = results[0].DisplayText;

                if (Regex.IsMatch(speechText, "(?i)(versende|sende).*?(email|e-mail).*?mein.*?gruppe"))
                {
                    this.WriteLine("Versende eine e-mail an meine Gruppe");
                    this.SendMail(mailList);
                }
                else if (Regex.IsMatch(speechText, "(?i)(versende|sende).*?(email|e-mail).*?(zwei|2).*?mitglieder"))
                {
                    this.WriteLine("Versende eine e-mail an 2 Gruppenmitglieder");
                    if (mailList.Count > 2) {
                        this.SendMail(mailList.GetRange(0, 2));
                    }
                    else {
                        this.WriteLine("Maillist enthält nicht genügend Mitglieder (maillist < 2)");
                    }

                }
                else
                {
                    this.WriteLine("** No match! Please retry.");
                }
            }
        }

        /// <summary>
        /// sends an email to the detected persons
        /// </summary>
        /// <param name="mailReceivers"></param>
        void SendMail(List<string> mailReceivers)
        {
            MailMessage mail = new MailMessage();
            mail.From = new MailAddress("spam@stedos.net");
            mail.Subject = "Face & Speech API - Application: You've been spotted!";
            mail.IsBodyHtml = true; //Nur wenn Body HTML Quellcode ist  
            string receiverStr = "";

            foreach (string actMail in mailReceivers)
            {
                mail.To.Add(actMail);
                receiverStr += " " + actMail;
            }
            mail.Body = "<h1>Face & Speech API - Application</h1><p> " + receiverStr + "</p>";

            try
            {
                SmtpClient client = new SmtpClient("smtp.1und1.de", 587);
                client.Credentials = new System.Net.NetworkCredential("spam@stedos.net", "JustForTesting_1");//Anmeldedaten für den SMTP Server 
                client.EnableSsl = true; //Die meisten Anbieter verlangen eine SSL-Verschlüsselung 
                client.Send(mail); //Senden 
                WriteLine("E-Mail wurde versendet an: {0}", receiverStr);
                mailList.Clear();
            }
            catch (Exception ex)
            {
                WriteLine("Fehler beim Senden der E-Mail\n\n{0}", ex.Message);
            }
        }

        /// <summary>
        ///     Called when a partial response is received.
        /// </summary>
        void OnPartialResponseReceivedHandler(object sender, PartialSpeechResponseEventArgs e)
        {
            this.WriteLine("*** Partial Result: \t {0}", e.PartialResult);
        }

        /// <summary>
        ///     Called when an error is received.
        /// </summary>
        void OnConversationErrorHandler(object sender, SpeechErrorEventArgs e)
        {
            this.Dispatcher.Invoke((Action)(() =>
            {
                this.speechButton.IsEnabled = true;
            }));

            this.WriteLine("********* Error Detected *********");
            this.WriteLine("{0}", e.SpeechErrorCode.ToString());
            this.WriteLine("{0}", e.SpeechErrorText);
        }

        /// <summary>
        ///     Called when the microphone status has changed.
        /// </summary>
        void OnMicrophoneStatus(object sender, MicrophoneEventArgs e)
        {
            if (!e.Recording)
            {
                m_micClient.EndMicAndRecognition();
                this.Dispatcher.Invoke((Action)(() =>
                {
                    this.speechButton.IsEnabled = true;
                }));
            }
            this.WriteLine("Microphone status: {0}", e.Recording);
        }

        /// <summary>s
        /// Writes the line.
        /// </summary>s
        /// <param name="format">The format.</param>
        /// <param name="args">The arguments.</param>
        void WriteLine(string format, params object[] args)
        {
            var formattedStr = string.Format(format, args);
            Trace.WriteLine(formattedStr);
            this.Dispatcher.Invoke(() =>
            {
                this.textBox.Text += (formattedStr + "\n");
            });
        }
    }
}