﻿// <copyright file="MainWindow.xaml.cs" company="Microsoft">
//     Copyright (c) Microsoft Corporation.  All rights reserved.
//     MIT License
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge, 
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons 
// to whom the Software is furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all copies or 
// substantial portions of the Software.
// THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// </copyright>
//----------------------------------------------------------------------

using MicrosoftProjectOxford;
using System;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Windows;
using System.Net.Mail;
using System.Collections;
using System.Collections.Generic; 

namespace MicrosoftProjectOxfordExample
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        string m_primaryOrSecondaryKey = ConfigurationManager.AppSettings["primaryKey"];

        private const string ShortWaveFile = @"emailsendengruppe.wav";

        private string m_filename = null;
        private bool m_isMicrophoneReco = true;
        private SpeechRecognitionMode m_recoMode = SpeechRecognitionMode.ShortPhrase;
        private MicrophoneRecognitionClient m_micClient;

        /// <summary>
        /// Initializes a new instance of the <see cref="MainWindow"/> class.
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();
            Intialize();
        }

        private void Intialize()
        {
                m_micClient = SpeechRecognitionServiceFactory.CreateMicrophoneClient(m_recoMode, "de-de", m_primaryOrSecondaryKey);
              
                // Event handlers for speech recognition results
                m_micClient.OnResponseReceived += this.OnResponseReceivedHandler;
                m_micClient.OnPartialResponseReceived += this.OnPartialResponseReceivedHandler;
                m_micClient.OnConversationError += this.OnConversationErrorHandler;
                m_micClient.OnMicrophoneStatus += this.OnMicrophoneStatus;
        }

        /// <summary>
        /// Handles the Click event of the button1 control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void button1_Click(object sender, RoutedEventArgs e)
        {
            this.editText1.Text = String.Empty;
            this.button1.IsEnabled = false;
            m_micClient.StartMicAndRecognition();
        }

        /// <summary>
        /// Handles the Click event of the button2 control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void button2_Click(object sender, RoutedEventArgs e)
        {
            bool isReceivedResponse = false;
            int waitSeconds = 15;

            if (null != m_micClient)
            {
                isReceivedResponse = m_micClient.WaitForFinalResponse(waitSeconds * 1000);
                // We are done sending audio. Final recognition results will arrive in OnResponseReceived event call.
                m_micClient.EndMicAndRecognition();
                this.WriteLine("App Done.");
                this.button1.IsEnabled = false;
                m_micClient.Dispose();
                m_micClient = null;
            }

            if (!isReceivedResponse)
            {
                this.button1.IsEnabled = false;
                this.WriteLine("{0}: Timed out waiting for conversation response after {1} ms",
                               DateTime.UtcNow, waitSeconds * 1000);
            }
        }

        /// <summary>
        ///     Called when a final response is received; 
        /// </summary>
        void OnResponseReceivedHandler(object sender, SpeechResponseEventArgs e)
        {
            //TODO: get this von face API / JSON
            List<string> mailList = new List<string>();
            mailList.Add("s_d@uni.de");
            mailList.Add("admin@stedos.net");
            mailList.Add("steven.dostert@uni.de");
            //mailList.Add("ederbenjamin@gmx.de");
            //mailList.Add("natalie-kurz1@gmx.de");
            //mailList.Add("jo.schuster@me.com");

            bool isFinalDicationMessage = e.PhraseResponse.RecognitionStatus == RecognitionStatus.EndOfDictation ||
                                           e.PhraseResponse.RecognitionStatus == RecognitionStatus.DictationEndSilenceTimeout;

            if (isFinalDicationMessage)
            {
                this.Dispatcher.Invoke((Action)(() =>
                {
                    this.button1.IsEnabled = true;
                }));
            }
            else
            {
                this.WriteLine("********* Final NBEST Results:: *********");
                for (int i = 0; i < e.PhraseResponse.Results.Length; i++)
                {
                    this.WriteLine("[{0}] Confidence={1} Text=\"{2}\"",
                                   i, e.PhraseResponse.Results[i].Confidence,
                                   e.PhraseResponse.Results[i].DisplayText);
                }
                this.WriteLine();

                /* **************************************
                 * Analyse speech text
                 * ************************************** */

                // Get first speech text result
                string speechText = e.PhraseResponse.Results[0].DisplayText;

                // Regex for "Versende eine eMail an meine Gruppe" (case insensitive)
                string regex1 = "(?i)(versende|sende).*?(email|e-mail).*?mein.*?gruppe";

                // Regex for "Versende eine eMail an 2 Gruppenmitglieder" (case insensitive)
                string regex2 = "(?i)(versende|sende).*?(email|e-mail).*?(zwei|2).*?mitglieder";

                if (System.Text.RegularExpressions.Regex.IsMatch(speechText, regex1))
                {
                    // Case "Versende eine eMail an meine Gruppe"
                    this.WriteLine("Versende eine eMail an meine Gruppe");
                    this.SendMail(mailList, mailList.Count);
                }
                else if (System.Text.RegularExpressions.Regex.IsMatch(speechText, regex2))
                {
                    // Case "Versende eine eMail an 2 Gruppenmitglieder"
                    this.WriteLine("Versende eine eMail an 2 Gruppenmitglieder");
                    this.SendMail(mailList, 2);
                }
                else
                {
                    this.WriteLine("--- No match! ---");
                }
            }
        }

        void SendMail(List<string> mailList, int count)
        {
            MailMessage mail = new MailMessage();
            mail.From = new MailAddress("spam@stedos.net");
            mail.Subject = "Face & Speech API - Application: You've been spotted!";
            mail.Body = "<h1>Face & Speech API - Application</h1>";
            mail.IsBodyHtml = true; //Nur wenn Body HTML Quellcode ist  

            foreach (string actMail in mailList.GetRange(0, count))
            {
                mail.To.Add(actMail);
            }

            try
            {
                SmtpClient client = new SmtpClient("smtp.1und1.de", 587);
                client.Credentials = new System.Net.NetworkCredential("spam@stedos.net", "JustForTesting_1");//Anmeldedaten für den SMTP Server 
                client.EnableSsl = true; //Die meisten Anbieter verlangen eine SSL-Verschlüsselung 
                client.Send(mail); //Senden 
                WriteLine("E-Mail wurde versendet");
            }
            catch (Exception ex)
            {
                WriteLine("Fehler beim Senden der E-Mail\n\n{0}", ex.Message);
            }
        }

        /// <summary>
        ///     Called when a partial response is received.
        /// </summary>
        void OnPartialResponseReceivedHandler(object sender, PartialSpeechResponseEventArgs e)
        {
            this.WriteLine("*** Partial Result: \t {0}", e.PartialResult);
        }

        /// <summary>
        ///     Called when an error is received.
        /// </summary>
        void OnConversationErrorHandler(object sender, SpeechErrorEventArgs e)
        {
            this.Dispatcher.Invoke((Action) (() => {
                this.button1.IsEnabled = true;
            }));

            this.WriteLine("********* Error Detected *********");
            this.WriteLine("{0}", e.SpeechErrorCode.ToString());
            this.WriteLine("{0}", e.SpeechErrorText);
            this.WriteLine();
        }

        /// <summary>
        ///     Called when the microphone status has changed.
        /// </summary>
        void OnMicrophoneStatus(object sender, MicrophoneEventArgs e)
        {
            if (!e.Recording) {
                this.Dispatcher.Invoke((Action) (() => {
                    this.button1.IsEnabled = true;
                }));
            }

            this.WriteLine("********* Microphone status: {0} *********", e.Recording);
            this.WriteLine();
        }

        /// <summary>
        /// Writes the line.
        /// </summary>
        void WriteLine()
        {
            this.WriteLine(string.Empty);
        }

        /// <summary>
        /// Writes the line.
        /// </summary>
        /// <param name="format">The format.</param>
        /// <param name="args">The arguments.</param>
        void WriteLine(string format, params object[] args)
        {
            var formattedStr = string.Format(format, args);
            Trace.WriteLine(formattedStr);
            this.Dispatcher.Invoke(() =>
            {
                this.editText1.Text += (formattedStr + "\n");
            });
        }
    }
}
